//__INSERT_LICENSE__
//$Id: ffburg.cpp,v 1.22 2007/01/30 19:03:44 mstorti Exp $

#include <stdio.h>
#include <string.h>
#include <vector>
#include <cassert>

#include <src/fem.h>
#include <src/texthash.h>
#include <src/getprop.h>
#include <src/utils.h>
#include <src/util2.h>
#include <src/fastmat2.h>

#include "burgers.h"

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---: 
void UGlobalAJac::init(int ndim_,int ndof_,int nel_) {
  ndim = ndim_;
  ndof = ndof_;
  nel = nel_;
  tmp2.resize(1,ndof).set(1.);
}

void UGlobalAJac::comp_flux(FastMat2 &flux,FastMat2 &U) {
  flux.prod(U,u,1,2);
}

void UGlobalAJac::
comp_A_jac_n(FastMat2 &A_jac_n, FastMat2 &normal) {
  tmp5.prod(u,normal,-1,-1);
  double un = tmp5.get();
  A_jac_n.eye(un);
}

void UGlobalAJac::
comp_A_grad_U(FastMat2 &A_grad_U,FastMat2 &grad_U) {
  A_grad_U.prod(u,grad_U,-1,-1,1);
}
  
void UGlobalAJac::
comp_A_grad_N(FastMat2 &A_grad_N,FastMat2 &dshapex) {
  tmp.prod(u,dshapex,-1,-1,1);
  for (int j=1; j<=nel; j++) {
    A_grad_N.ir(1,j).eye(tmp.get(j));
  }
  A_grad_N.rs();
}
  
void UGlobalAJac::
comp_Uintri(FastMat2 &Uintri,FastMat2 &iJaco) {
  tmp3.prod(iJaco,u,1,-1,-1);
  Uintri.prod(tmp2,tmp3,1,2);
}

void UGlobalAJac::
comp_vel_per_field(FastMat2 &vel_per_field) {
  double vv= u.sum_square_all();
  vel_per_field.set(sqrt(vv));
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---: 
void GlobalScalarDJac::init(int ndim_,int ndof_,int nel_) {
  ndim = ndim_;
  ndof = ndof_;
  nel = nel_;
  eye_ndof.resize(2,ndof,ndof).eye(1.);
}

void GlobalScalarDJac
::comp_fluxd(FastMat2 &fluxd,FastMat2 &grad_U) {
  fluxd.t().set(grad_U).scale(diff).rs();
}

void GlobalScalarDJac
::comp_grad_N_D_grad_N(FastMat2 &grad_N_D_grad_N,
		       FastMat2 & dshapex,double w) {
  tmp.prod(dshapex,dshapex,-1,1,-1,2).scale(w*diff);
  grad_N_D_grad_N.prod(tmp,eye_ndof,1,3,2,4);
}

void GlobalScalarDJac
::comp_dif_per_field(FastMat2 &dif_per_field) {
  dif_per_field.set(diff);
}  

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---: 
advdif_wjac_ff::advdif_wjac_ff(NewElemset *elemset_,
			       DJac *d,AJac *a, 
			       CJac *c, SourceTerm *st) 
  : NewAdvDifFF(elemset_), 
    a_jac(a), d_jac(d), c_jac(c),
    source_term(st) { }

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---: 
#if 0
#undef __FUNC__
#define __FUNC__ "burgers_ff::burgers_ff()"
burgers_ff::burgers_ff(NewAdvDif *elemset_) 
  : advdif_wjac_ff(elemset_), a_jac(&u_a_jac),
    source_term(&st), c_jac(&n_c_jac), d_jac(&g_d_jac) { }
#endif

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---: 
void burgers_ff::start_chunk(int &ret_options) {
  new_adv_dif_elemset = dynamic_cast<const NewAdvDif *>(elemset); 
  FastMat2 &u = u_a_jac.u;
  int ierr;
  EGETOPTDEF_ND(elemset,int,ndim,0); //nd

  //o Scale the SUPG upwind term. Set to 0 in order to
  //  not to include the upwind term. 
  EGETOPTDEF_ND(elemset,double,tau_fac,1.);
  //o Add shock-capturing term.
  EGETOPTDEF_ND(elemset,int,shock_capturing,0);
  //o Add shock-capturing factor term.
  EGETOPTDEF_ND(elemset,double,shocap_fac,1);
  //o Add shock-capturing exponent.
  EGETOPTDEF_ND(elemset,double,shocap_beta,1);

  elemset->elem_params(nel,ndof,nelprops);

  enthalpy_fun = &identity_ef;
  identity_ef.init(ndof,ndim,nel);

  u_a_jac.init(ndim,ndof,nel);
  g_d_jac.init(ndim,ndof,nel);
  // this must be done here. Because this matrices may have
  // differente sizes for different elemsets. 
  u0.resize(1,ndim);
  u.resize(1,ndim);
  Uintri.resize(1,ndim);
  tmp0.resize(1,ndim);
  Cp.resize(2,ndof,ndof);
  grad_phi.resize(1,ndof);
  assert(ndof==1); // Only Burgers with one field is considered

  //o Diffusivity (viscosity)
  EGETOPTDEF_ND(elemset,double,diffusivity,0.);
  g_d_jac.diff = diffusivity;

  //o Flux law is $f= 0.5\,c\, \phi^2\,!u_0$, where $c$
  // is this coefficient. 
  EGETOPTDEF_ND(elemset,double,flux_law_coefficient,1.);
  //o Scale the SUPG upwind term. 
  EGETOPTDEF_ND(elemset,double,tau_fac,1.);

  //o _T: double[ndim] _N: u0 _D: [1,0] 
  // _DOC: Vector defining direction for flux. 
  // _END
  u0.set(0.); u0.setel(1.,1);
  ierr = elemset->get_double("u0",*u0.storage_begin(),1,ndim); 
  PETSCFEM_ASSERT0(ierr==0,"Error reading variable u0");
  ret_options |= SCALAR_TAU;	// tell the advective element routine
				// that we are returning a scalar tau
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>---: 
#undef __FUNC__
#define __FUNC__ "void burgers_ff::compute_flux(...)"
void burgers_ff::compute_flux(COMPUTE_FLUX_ARGS) {
  FastMat2 &u = u_a_jac.u;

  double coef = flux_law_coefficient;
  double phi = U.get(1);
  u.set(u0).scale(coef*phi);
  double vel = sqrt(u.sum_square_all());
  Ucpy.set(U);

  // A_jac.ir(2,1).ir(3,1).set(u).rs();
  flux.ir(1,1).set(u0).scale(0.5*coef*phi*phi).rs();

  if (options & COMP_UPWIND) {

#if 0
    D_jac.set(0.).ir(3,1).ir(4,1).eye(diffusivity).rs();
    grad_U.t();
    fluxd.set(grad_U).scale(diffusivity);
    grad_U.rs();
    A_grad_U.prod(A_jac.rs(),grad_U,-1,1,-2,-1,-2);
#endif

    d_jac->comp_fluxd(fluxd,grad_U);
    a_jac->comp_A_grad_U(A_grad_U,grad_U);

    lam_max = vel;
    tau_supg.set(0.);

    Uintri.prod(iJaco,u,1,-1,-1);
    double Uh = sqrt(Uintri.sum_square_all()); // this is
				// approx. 2*U/h

    double &alpha = diffusivity;
    double tau;
    double Pe  = vel*vel/(Uh*alpha),
      h = 2./sqrt(tmp0.sum_square(iJaco,1,-1).max_all()); // h and Peclet number
    FastMat2::branch();
    if (vel*vel > 1e-5*Uh*alpha) { // remove singularity when v=0
      FastMat2::choose(0);
      // magic function
      double magic = (fabs(Pe)>1.e-4 ? 1./tanh(Pe)-1./Pe : Pe/3.); 
      tau = tau_fac/Uh*magic; // intrinsic time
    } else {
      FastMat2::choose(1);
      tau = tau_fac*h*h/(12.*alpha);
    }
    FastMat2::leave();
    tau_supg.setel(tau,1,1);
    if (shock_capturing) {
      double grad_phi_mod = sqrt(grad_phi.sum_square_all());
      double fz = grad_phi_mod*h;
      fz = pow(fz,shocap_beta);
      
      double fz2 = (Pe < 3. ? Pe/3. : 1.);
      fz2 = pow(fz2,shocap_beta);
      delta_sc = 0.5*h*vel*fz2*shocap_fac;
      delta_sc_aniso = 0.5*h*vel*fz;
    } else {
      delta_sc = 0.;
    }
  }
  
  if (options & COMP_SOURCE) {
    G_source.set(0.);		// Only null source term allowed
				// right now!!
  }
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
void burgers_ff::get_Ajac(FastMat2 &Ajac) {
  FastMat2 &u = u_a_jac.u;
  Ajac.ir(2,1).ir(3,1).set(u).rs();
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
void burgers_ff::get_Cp(FastMat2 &Cp_a) {
  identity_ef.get_Cp(Cp_a);
  Cp.set(Cp_a).rs();
}
